#!/usr/bin/env python3

# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""This is a usage example of aiocoap that demonstrates how to implement a
simple server. See the "Usage Examples" section in the aiocoap documentation
for some more information."""

import datetime
import logging

import asyncio

import aiocoap.resource as resource
import aiocoap

class CoralDocument(resource.Resource):
    async def render_get(self, request):
        return aiocoap.Message(payload=open('/home/chrysn/git/micrurus/coffeemachine.coral', 'rb').read(), content_format=65087)
        #return aiocoap.Message(payload=open('/home/chrysn/git/micrurus/coffeemachine.yoral', 'rb').read(), content_format=65535)

class MenuDocument(resource.Resource):
    async def render_get(self, request):
        return aiocoap.Message(payload=b'{}', content_format=65502)

class BrewingResource(resource.Resource):
    async def render_post(self, request):
        return aiocoap.Message(location_path=["activity", "42"])

class BrewingTask(resource.ObservableResource):
    @staticmethod
    def represent_status(name):
        return aiocoap.Message(code=aiocoap.CONTENT, payload=('!coral:tag:chrysn@fsfe.org,2018-11-02:linkformat [[2, "http://coreapps.org/htcpc#status", "%s"]]' % name).encode('utf8'), content_format=65535)

    async def render_get(self, request):
        return self.represent_status("queued")

class BlockResource(resource.Resource):
    """Example resource which supports the GET and PUT methods. It sends large
    responses, which trigger blockwise transfer."""

    def __init__(self):
        super().__init__()
        self.set_content(b"This is the resource's default content. It is padded "
                b"with numbers to be large enough to trigger blockwise "
                b"transfer.\n")

    def set_content(self, content):
        self.content = content
        while len(self.content) <= 1024:
            self.content = self.content + b"0123456789\n"

    async def render_get(self, request):
        return aiocoap.Message(payload=self.content)

    async def render_put(self, request):
        print('PUT payload: %s' % request.payload)
        self.set_content(request.payload)
        return aiocoap.Message(code=aiocoap.CHANGED, payload=self.content)


class SeparateLargeResource(resource.Resource):
    """Example resource which supports the GET method. It uses asyncio.sleep to
    simulate a long-running operation, and thus forces the protocol to send
    empty ACK first. """

    def get_link_description(self):
        # Publish additional data in .well-known/core
        return dict(**super().get_link_description(), title="A large resource")

    async def render_get(self, request):
        await asyncio.sleep(3)

        payload = "Three rings for the elven kings under the sky, seven rings "\
                "for dwarven lords in their halls of stone, nine rings for "\
                "mortal men doomed to die, one ring for the dark lord on his "\
                "dark throne.".encode('ascii')
        return aiocoap.Message(payload=payload)

class TimeResource(resource.ObservableResource):
    """Example resource that can be observed. The `notify` method keeps
    scheduling itself, and calles `update_state` to trigger sending
    notifications."""

    def __init__(self):
        super().__init__()

        self.handle = None

    def notify(self):
        self.updated_state()
        self.reschedule()

    def reschedule(self):
        self.handle = asyncio.get_event_loop().call_later(5, self.notify)

    def update_observation_count(self, count):
        if count and self.handle is None:
            print("Starting the clock")
            self.reschedule()
        if count == 0 and self.handle:
            print("Stopping the clock")
            self.handle.cancel()
            self.handle = None

    async def render_get(self, request):
        payload = datetime.datetime.now().\
                strftime("%Y-%m-%d %H:%M").encode('ascii')
        return aiocoap.Message(payload=payload)

# logging setup

logging.basicConfig(level=logging.INFO)
logging.getLogger("coap-server").setLevel(logging.DEBUG)

def main():
    # Resource tree creation
    root = resource.Site()

    root.add_resource(('.well-known', 'core'),
            resource.WKCResource(root.get_resources_as_linkheader))
    root.add_resource(('time',), TimeResource())
    root.add_resource(('other', 'block'), BlockResource())
    root.add_resource(('other', 'separate'), SeparateLargeResource())
    root.add_resource((), CoralDocument())
    root.add_resource(['menu'], MenuDocument())
    root.add_resource(['brew'], BrewingResource())

    brewtask = BrewingTask()
    root.add_resource(['activity', '42'], brewtask)

    asyncio.Task(aiocoap.Context.create_server_context(root))

    async def cycling():
        while True:
            await asyncio.sleep(2)
            brewtask.updated_state(BrewingTask.represent_status("queued"))
            await asyncio.sleep(2)
            brewtask.updated_state(BrewingTask.represent_status("brewing"))
            await asyncio.sleep(2)
            brewtask.updated_state(BrewingTask.represent_status("closed"))

    asyncio.get_event_loop().run_until_complete(cycling())

if __name__ == "__main__":
    main()
